(defvar magic-square '(2 9 4 7 5 3 6 1 8))

; States will be lists of two lists, e.g., ((1 2 3) (4 5 6)), where the first list
; is the location of the X's and the second list is the location of the O's.
; The locations refer not to the standard positions, but to the "magic square" positions:
;
;    2 9 4
;    7 5 3
;    6 1 8
;
; Labelling the squares of the Tic-Tac-Toe board this way is useful because then I can
; just add up any three positions, and if the sum is 15, then I know they are three in
; a row.  The following function then tells me if a list of X or O positions contains
; any that are three in a row.

(defun any-n-sum-to-k? (n k list)
  (cond ((= n 0)
         (= k 0))
        ((< k 0)
         nil)
        ((null list)
         nil)
        ((any-n-sum-to-k? (- n 1) (- k (first list)) (rest list))
         t)                             ; either the first element is included
        ((any-n-sum-to-k? n k (rest list))
         t)))                           ; or it's not

(defvar value-table)
(defvar initial-state)

#|
; The value function will be implemented as a hash table.  This has two advantages.
; First, I can acess it with states in the two-list format without converting to
; an array index.  Second, I won't allocate space for all those states that never occur.

(defun init ()
  (setq initial-state '(nil nil))
  (setq value-table (make-hash-table :test #'equal)))

(defun value (state)
  (or (gethash state value-table)
      (setf (gethash state value-table)
            (cond ((any-n-sum-to-k? 3 15 (first state))
                   0)
                  ((any-n-sum-to-k? 3 15 (second state))
                   1)
                  ((= 9 (+ (length (first state)) (length (second state))))
                   0)
                  (t 0.5)))))

(defun set-value (state value)
  (setf (gethash state value-table)
        value))

(defun next-state (player state move)
  "returns new state after making the indicated move by the indicated player (:X or :O)"
  (let ((X-moves (first state))
        (O-moves (second state)))
    (if move
      (list (if (eq player :X)
              (sort (copy-list (cons move X-moves)) #'>)
              X-moves)
            (if (eq player :O)
              (sort (copy-list (cons move O-moves)) #'>)
              O-moves))
      state)))

|#

; The value function will be implemented as a big, mostly empty array.  A state will be of the
; form (X-locations O-locations index), where the index is an index into the big array.
; The index is computed from the locations, which are unordered sets.

(defun state-index (X-locations O-locations)
  (+ (loop for l in X-locations sum (aref powers-of-2 l))
     (* 512 (loop for l in O-locations sum (aref powers-of-2 l)))))

(defvar powers-of-2 (make-array 10 :initial-contents (cons nil (loop for i below 9 collect (expt 2 i)))))

(defun init ()
  (setq value-table (make-array (* 512 512) :initial-element nil))
  (setq initial-state '(nil nil 0))
  (set-value initial-state 0.5)
  (values))

(defun value (state)
  (aref value-table (third state)))

(defun set-value (state value)
  (setf (aref value-table (third state)) value))
  
(defun next-state (player state move)
  "returns new state after making the indicated move by the indicated player (:X or :O)"
  (let ((X-moves (first state))
        (O-moves (second state)))
    (if (eq player :X)
      (push move X-moves)
      (push move O-moves))
    (setq state (list X-moves O-moves (state-index X-moves O-moves)))
    (when (null (value state))
      (set-value state (cond ((any-n-sum-to-k? 3 15 X-moves)
                              0)
                             ((any-n-sum-to-k? 3 15 O-moves)
                              1)
                             ((= 9 (+ (length X-moves) (length O-moves)))
                              0)
                             (t 0.5))))
    state))


(defun terminal-state-p (state)
  (integerp (value state)))

(defvar alpha 0.1)

(defun random-move (state)
  (let ((possible-moves (possible-moves state)))
    (if (null possible-moves)
      nil
      (nth (random (length possible-moves))
           possible-moves))))

(defun greedy-move (player state)
  (let ((possible-moves (possible-moves state)))
    (if (null possible-moves)
      nil
      (loop with best-value = -1
            with best-move
            for move in possible-moves
            for move-value = (value (next-state player state move))
            do (when (> move-value best-value) 
                 (setf best-value move-value)
                 (setf best-move move))
            finally (return best-move)))))

; Now here is the main function

(defvar state)

(defun game (&optional quiet)
  (setq state initial-state)
  (unless quiet (show-state state))
  (loop for new-state = (next-state :X state (random-move state)) 
        for exploratory-move? = (= 0 (random 100000))
        do
        (when (terminal-state-p new-state)
          (unless quiet (show-state new-state))
          (update state new-state quiet)
          (return (value new-state)))
        (setf new-state (next-state :O new-state 
                                    (if exploratory-move?
                                      (random-move new-state)
                                      (greedy-move :O new-state))))
        (unless exploratory-move?
          (update state new-state quiet))
        (unless quiet (show-state new-state))
        (when (terminal-state-p new-state) (return (value new-state)))
        (setq state new-state)))

(defun update (state new-state &optional quiet)
  (set-value state (+ (value state)
                      (* alpha
                         (- (value new-state)
                            (value state)))))
  (unless quiet (format t "                    ~,3F" (value state))))

(defun possible-moves (state)
  (loop for i from 1 to 9 
        unless (or (member i (first state))
                   (member i (second state)))
        collect i))

(defun show-state (state)
  (let ((X-moves (first state))
        (O-moves (second state)))
    (format t "~%")
    (loop for location in magic-square
          for i from 0
          do
          (format t (cond ((member location X-moves)
                           " X")
                          ((member location O-moves)
                           " O")
                          (t " -")))
          (when (= i 5) (format t "  ~,3F" (value state)))
          (when (= 2 (mod i 3)) (format t "~%"))))
  (values))
                 
